@@ -2,7 +2,7 @@
 #
 # The MIT License (MIT)
 #
-# Copyright (c) Tavendo GmbH
+# Copyright (c) Crossbar.io Technologies GmbH
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -25,46 +25,29 @@
 ###############################################################################
 
 from __future__ import absolute_import
+from __future__ import print_function
 
-import os
 import binascii
 import struct
 
 import six
-
-from txaio import create_future_success
+import txaio
 
 from autobahn import util
 from autobahn.wamp.types import Challenge
 
-from twisted.internet.defer import inlineCallbacks, returnValue
-
 __all__ = [
     'HAS_CRYPTOSIGN',
-    'HAS_CRYPTOSIGN_SSHAGENT'
 ]
 
 try:
     # try to import everything we need for WAMP-cryptosign
-    from nacl import public, encoding, signing, bindings
+    from nacl import encoding, signing, bindings
 except ImportError:
     HAS_CRYPTOSIGN = False
-    HAS_CRYPTOSIGN_SSHAGENT = False
 else:
     HAS_CRYPTOSIGN = True
     __all__.append('SigningKey')
-    try:
-        # WAMP-cryptosign support for SSH agent is currently
-        # only available on Twisted (on Python 2)
-        from twisted.internet.protocol import Factory
-        from twisted.internet.endpoints import UNIXClientEndpoint
-        from twisted.conch.ssh.agent import SSHAgentClient
-    except ImportError:
-        # twisted.conch is not yet fully ported to Python 3
-        HAS_CRYPTOSIGN_SSHAGENT = False
-    else:
-        HAS_CRYPTOSIGN_SSHAGENT = True
-        __all__.append('SSHAgentSigningKey')
 
 
 def _unpack(keydata):
@@ -104,7 +87,7 @@ def _read_ssh_ed25519_pubkey(keydata):
         ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIJukDU5fqXv/yVhSirsDWsUFyOodZyCSLxyitPPzWJW9 oberstet@office-corei7
 
     :param keydata: The OpenSSH Ed25519 public key data to parse.
-    :type keydata: unicode
+    :type keydata: str
 
     :returns: pair of raw public key (32 bytes) and comment
     :rtype: tuple
@@ -155,12 +138,16 @@ class _SSHPacketReader:
         return value
 
     def get_uint32(self):
-        return int.from_bytes(self.get_bytes(4), 'big')
+        return struct.unpack('>I', self.get_bytes(4))[0]
 
     def get_string(self):
         return self.get_bytes(self.get_uint32())
 
 
+def _makepad(size):
+    return ''.join(chr(x) for x in range(1, size + 1))
+
+
 def _read_ssh_ed25519_privkey(keydata):
     """
     Parse an OpenSSH Ed25519 private key from a string into a raw private key.
@@ -177,7 +164,7 @@ def _read_ssh_ed25519_privkey(keydata):
 
 
     :param keydata: The OpenSSH Ed25519 private key data to parse.
-    :type keydata: unicode
+    :type keydata: str
 
     :returns: pair of raw private key (32 bytes) and comment
     :rtype: tuple
@@ -250,8 +237,8 @@ def _read_ssh_ed25519_privkey(keydata):
     comment = packet.get_string()                             # comment
     pad = packet.get_remaining_payload()
 
-    if len(pad) >= block_size or pad != bytes(range(1, len(pad) + 1)):
-        raise Exception('invalid OpenSSH private key')
+    if len(pad) and (len(pad) >= block_size or pad != _makepad(len(pad))):
+        raise Exception('invalid OpenSSH private key (padlen={}, actual_pad={}, expected_pad={})'.format(len(pad), pad, _makepad(len(pad))))
 
     # secret key (64 octets) = 32 octets seed || 32 octets secret key derived of seed
     seed = sk[:bindings.crypto_sign_SEEDBYTES]
@@ -370,6 +357,7 @@ def _verify_signify_ed25519_signature(pubkey_file, signature_file, message):
 
 if HAS_CRYPTOSIGN:
 
+    @util.public
     class SigningKey(object):
         """
         A cryptosign private key for signing, and hence usable for authentication or a
@@ -380,7 +368,7 @@ if HAS_CRYPTOSIGN:
             """
 
             :param key: A Ed25519 private signing key or a Ed25519 public verification key.
-            :type key: instance of nacl.public.VerifyKey or instance of nacl.public.SigningKey
+            :type key: instance of nacl.signing.VerifyKey or instance of nacl.signing.SigningKey
             """
             if not (isinstance(key, signing.VerifyKey) or isinstance(key, signing.SigningKey)):
                 raise Exception("invalid type {} for key".format(type(key)))
@@ -393,8 +381,10 @@ if HAS_CRYPTOSIGN:
             self._can_sign = isinstance(key, signing.SigningKey)
 
         def __str__(self):
-            return u'Key(can_sign={}, comment="{}", public_key={})'.format(self.can_sign(), self.comment(), self.public_key())
+            comment = u'"{}"'.format(self.comment()) if self.comment() else None
+            return u'Key(can_sign={}, comment={}, public_key={})'.format(self.can_sign(), comment, self.public_key())
 
+        @util.public
         def can_sign(self):
             """
             Check if the key can be used to sign.
@@ -404,21 +394,23 @@ if HAS_CRYPTOSIGN:
             """
             return self._can_sign
 
+        @util.public
         def comment(self):
             """
             Get the key comment (if any).
 
             :returns: The comment (if any) from the key.
-            :rtype: unicode or None
+            :rtype: str or None
             """
             return self._comment
 
+        @util.public
         def public_key(self, binary=False):
             """
             Returns the public key part of a signing key or the (public) verification key.
 
             :returns: The public key in Hex encoding.
-            :rtype: unicode or None
+            :rtype: str or None
             """
             if isinstance(self._key, signing.SigningKey):
                 key = self._key.verify_key
@@ -430,6 +422,7 @@ if HAS_CRYPTOSIGN:
             else:
                 return key.encode(encoder=encoding.HexEncoder).decode('ascii')
 
+        @util.public
         def sign(self, data):
             """
             Sign some data.
@@ -453,9 +446,9 @@ if HAS_CRYPTOSIGN:
             # it get coerced into the concatenation of message + signature
             # not sure which order, but we don't want that. we only want
             # the signature
-            return create_future_success(sig.signature)
+            return txaio.create_future_success(sig.signature)
 
-        @inlineCallbacks
+        @util.public
         def sign_challenge(self, session, challenge):
             """
             Sign WAMP-cryptosign challenge.
@@ -464,7 +457,7 @@ if HAS_CRYPTOSIGN:
             :type challenge: instance of autobahn.wamp.types.Challenge
 
             :returns: A Deferred/Future that resolves to the computed signature.
-            :rtype: unicode
+            :rtype: str
             """
             if not isinstance(challenge, Challenge):
                 raise Exception("challenge must be instance of autobahn.wamp.types.Challenge, not {}".format(type(challenge)))
@@ -487,16 +480,37 @@ if HAS_CRYPTOSIGN:
                 data = challenge_raw
 
             # a raw byte string is signed, and the signature is also a raw byte string
-            signature_raw = yield self.sign(data)
+            d1 = self.sign(data)
 
-            # convert the raw signature into a hex encode value (unicode string)
-            signature_hex = binascii.b2a_hex(signature_raw).decode('ascii')
+            # asyncio lacks callback chaining (and we cannot use co-routines, since we want
+            # to support older Pythons), hence we need d2
+            d2 = txaio.create_future()
 
-            # we return the concatenation of the signature and the message signed (96 bytes)
-            data_hex = binascii.b2a_hex(data).decode('ascii')
+            def process(signature_raw):
+                # convert the raw signature into a hex encode value (unicode string)
+                signature_hex = binascii.b2a_hex(signature_raw).decode('ascii')
 
-            # we always return a future/deferred, so handling is uniform
-            returnValue(signature_hex + data_hex)
+                # we return the concatenation of the signature and the message signed (96 bytes)
+                data_hex = binascii.b2a_hex(data).decode('ascii')
+
+                sig = signature_hex + data_hex
+                txaio.resolve(d2, sig)
+
+            txaio.add_callbacks(d1, process, None)
+
+            return d2
+
+        @util.public
+        @classmethod
+        def from_key_bytes(cls, keydata, comment=None):
+            if not (comment is None or type(comment) == six.text_type):
+                raise ValueError("invalid type {} for comment".format(type(comment)))
+
+            if len(keydata) != 32:
+                raise ValueError("invalid key length {}".format(len(keydata)))
+
+            key = signing.SigningKey(keydata)
+            return cls(key, comment)
 
         @classmethod
         def from_raw_key(cls, filename, comment=None):
@@ -511,9 +525,9 @@ if HAS_CRYPTOSIGN:
                 dd if=/dev/urandom of=client02.key bs=1 count=32
 
             :param filename: Filename of the key.
-            :type filename: unicode
+            :type filename: str
             :param comment: Comment for key (optional).
-            :type comment: unicode or None
+            :type comment: str or None
             """
             if not (comment is None or type(comment) == six.text_type):
                 raise Exception("invalid type {} for comment".format(type(comment)))
@@ -524,12 +538,9 @@ if HAS_CRYPTOSIGN:
             with open(filename, 'rb') as f:
                 keydata = f.read()
 
-            if len(keydata) != 32:
-                raise Exception("invalid key length {}".format(len(keydata)))
-
-            key = signing.SigningKey(keydata)
-            return cls(key, comment)
+            return cls.from_key_bytes(keydata, comment=comment)
 
+        @util.public
         @classmethod
         def from_ssh_key(cls, filename):
             """
@@ -537,118 +548,51 @@ if HAS_CRYPTOSIGN:
             key (from a SSH private key file) or a (public) verification key (from a SSH
             public key file). A private key file must be passphrase-less.
             """
+
+            with open(filename, 'rb') as f:
+                keydata = f.read().decode('utf-8').strip()
+            return cls.from_ssh_data(keydata)
+
+        @util.public
+        @classmethod
+        def from_ssh_data(cls, keydata):
+            """
+            Load an Ed25519 key from SSH key file. The key file can be a (private) signing
+            key (from a SSH private key file) or a (public) verification key (from a SSH
+            public key file). A private key file must be passphrase-less.
+            """
             SSH_BEGIN = u'-----BEGIN OPENSSH PRIVATE KEY-----'
-
-            with open(filename, 'r') as f:
-                keydata = f.read().strip()
-
             if keydata.startswith(SSH_BEGIN):
                 # OpenSSH private key
                 keydata, comment = _read_ssh_ed25519_privkey(keydata)
                 key = signing.SigningKey(keydata, encoder=encoding.RawEncoder)
             else:
                 # OpenSSH public key
-                keydata, comment = _read_ssh_ed25519_pubkey(filename)
-                key = public.PublicKey(keydata, encoder=encoding.RawEncoder)
+                keydata, comment = _read_ssh_ed25519_pubkey(keydata)
+                key = signing.VerifyKey(keydata)
 
             return cls(key, comment)
 
+if __name__ == '__main__':
+    import sys
+    if not HAS_CRYPTOSIGN:
+        print('NaCl library must be installed for this to function.', file=sys.stderr)
+        sys.exit(1)
 
-if HAS_CRYPTOSIGN_SSHAGENT:
+    from optparse import OptionParser
 
-    class SSHAgentSigningKey(SigningKey):
-        """
-        A WAMP-cryptosign signing key that is a proxy to a private Ed25510 key
-        actually held in SSH agent.
+    parser = OptionParser()
+    parser.add_option('-f', '--file', dest='keyfile',
+                      help='file containing ssh key')
+    parser.add_option('-p', action='store_true', dest='printpub', default=False,
+                      help='print public key information')
 
-        An instance of this class must be create via the class method new().
-        The instance only holds the public key part, whereas the private key
-        counterpart is held in SSH agent.
-        """
+    options, args = parser.parse_args()
 
-        def __init__(self, key, comment=None, reactor=None):
-            SigningKey.__init__(self, key, comment)
-            if not reactor:
-                from twisted.internet import reactor
-            self._reactor = reactor
+    if not options.printpub:
+        print("Print public key must be specified as it's the only option.")
+        parser.print_usage()
+        sys.exit(1)
 
-        @classmethod
-        def new(cls, pubkey=None, reactor=None):
-            """
-            Create a proxy for a key held in SSH agent.
-
-            :param pubkey: A string with a public Ed25519 key in SSH format.
-            :type pubkey: unicode
-            """
-            if not HAS_CRYPTOSIGN_SSHAGENT:
-                raise Exception("SSH agent integration is not supported on this platform")
-
-            pubkey, _ = _read_ssh_ed25519_pubkey(pubkey)
-
-            if not reactor:
-                from twisted.internet import reactor
-
-            if "SSH_AUTH_SOCK" not in os.environ:
-                raise Exception("no ssh-agent is running!")
-
-            factory = Factory()
-            factory.noisy = False
-            factory.protocol = SSHAgentClient
-            endpoint = UNIXClientEndpoint(reactor, os.environ["SSH_AUTH_SOCK"])
-            d = endpoint.connect(factory)
-
-            @inlineCallbacks
-            def on_connect(agent):
-                keys = yield agent.requestIdentities()
-
-                # if the key is found in ssh-agent, the raw public key (32 bytes), and the
-                # key comment as returned from ssh-agent
-                key_data = None
-                key_comment = None
-
-                for blob, comment in keys:
-                    raw = _unpack(blob)
-                    algo = raw[0]
-                    if algo == u'ssh-ed25519':
-                        algo, _pubkey = raw
-                        if _pubkey == pubkey:
-                            key_data = _pubkey
-                            key_comment = comment.decode('utf8')
-                            break
-
-                agent.transport.loseConnection()
-
-                if key_data:
-                    key = signing.VerifyKey(key_data)
-                    returnValue(cls(key, key_comment, reactor))
-                else:
-                    raise Exception("Ed25519 key not held in ssh-agent")
-
-            return d.addCallback(on_connect)
-
-        def sign(self, challenge):
-            if "SSH_AUTH_SOCK" not in os.environ:
-                raise Exception("no ssh-agent is running!")
-
-            factory = Factory()
-            factory.noisy = False
-            factory.protocol = SSHAgentClient
-            endpoint = UNIXClientEndpoint(self._reactor, os.environ["SSH_AUTH_SOCK"])
-            d = endpoint.connect(factory)
-
-            @inlineCallbacks
-            def on_connect(agent):
-                # we are now connected to the locally running ssh-agent
-                # that agent might be the openssh-agent, or eg on Ubuntu 14.04 by
-                # default the gnome-keyring / ssh-askpass-gnome application
-                blob = _pack(['ssh-ed25519', self.public_key(binary=True)])
-
-                # now ask the agent
-                signature_blob = yield agent.signData(blob, challenge)
-                algo, signature = _unpack(signature_blob)
-
-                agent.transport.loseConnection()
-
-                returnValue(signature)
-
-            return d.addCallback(on_connect)
+    key = SigningKey.from_ssh_key(options.keyfile)
+    print(key.public_key())
